
cdef extern from "<cmath>" namespace "std" nogil:
    # all C99 functions
    float acos(float x) except +
    double acos(double x) except +
    long double acos(long double x) except +
    float acosf(float x) except +
    long double acosl(long double x) except +

    float asin(float x) except +
    double asin(double x) except +
    long double asin(long double x) except +
    float asinf(float x) except +
    long double asinl(long double x) except +

    float atan(float x) except +
    double atan(double x) except +
    long double atan(long double x) except +
    float atanf(float x) except +
    long double atanl(long double x) except +

    float atan2(float y, float x) except +
    double atan2(double y, double x) except +
    long double atan2(long double y, long double x) except +
    float atan2f(float y, float x) except +
    long double atan2l(long double y, long double x) except +

    float cos(float x) except +
    double cos(double x) except +
    long double cos(long double x) except +
    float cosf(float x) except +
    long double cosl(long double x) except +

    float sin(float x) except +
    double sin(double x) except +
    long double sin(long double x) except +
    float sinf(float x) except +
    long double sinl(long double x) except +

    float tan(float x) except +
    double tan(double x) except +
    long double tan(long double x) except +
    float tanf(float x) except +
    long double tanl(long double x) except +

    float acosh(float x) except +
    double acosh(double x) except +
    long double acosh(long double x) except +
    float acoshf(float x) except +
    long double acoshl(long double x) except +

    float asinh(float x) except +
    double asinh(double x) except +
    long double asinh(long double x) except +
    float asinhf(float x) except +
    long double asinhl(long double x) except +

    float atanh(float x) except +
    double atanh(double x) except +
    long double atanh(long double x) except +
    float atanhf(float x) except +
    long double atanhl(long double x) except +

    float cosh(float x) except +
    double cosh(double x) except +
    long double cosh(long double x) except +
    float coshf(float x) except +
    long double coshl(long double x) except +

    float sinh(float x) except +
    double sinh(double x) except +
    long double sinh(long double x) except +
    float sinhf(float x) except +
    long double sinhl(long double x) except +

    float tanh(float x) except +
    double tanh(double x) except +
    long double tanh(long double x) except +
    float tanhf(float x) except +
    long double tanhl(long double x) except +

    float exp(float x) except +
    double exp(double x) except +
    long double exp(long double x) except +
    float expf(float x) except +
    long double expl(long double x) except +

    float exp2(float x) except +
    double exp2(double x) except +
    long double exp2(long double x) except +
    float exp2f(float x) except +
    long double exp2l(long double x) except +

    float expm1(float x) except +
    double expm1(double x) except +
    long double expm1(long double x) except +
    float expm1f(float x) except +
    long double expm1l(long double x) except +

    float frexp(float value, int* exp) except +
    double frexp(double value, int* exp) except +
    long double frexp(long double value, int* exp) except +
    float frexpf(float value, int* exp) except +
    long double frexpl(long double value, int* exp) except +

    int ilogb(float x) except +
    int ilogb(double x) except +
    int ilogb(long double x) except +
    int ilogbf(float x) except +
    int ilogbl(long double x) except +

    float ldexp(float x, int exp) except +
    double ldexp(double x, int exp) except +
    long double ldexp(long double x, int exp) except +
    float ldexpf(float x, int exp) except +
    long double ldexpl(long double x, int exp) except +

    float log(float x) except +
    double log(double x) except +
    long double log(long double x) except +
    float logf(float x) except +
    long double logl(long double x) except +

    float log10(float x) except +
    double log10(double x) except +
    long double log10(long double x) except +
    float log10f(float x) except +
    long double log10l(long double x) except +

    float log1p(float x) except +
    double log1p(double x) except +
    long double log1p(long double x) except +
    float log1pf(float x) except +
    long double log1pl(long double x) except +

    float log2(float x) except +
    double log2(double x) except +
    long double log2(long double x) except +
    float log2f(float x) except +
    long double log2l(long double x) except +

    float logb(float x) except +
    double logb(double x) except +
    long double logb(long double x) except +
    float logbf(float x) except +
    long double logbl(long double x) except +

    float modf(float value, float* iptr) except +
    double modf(double value, double* iptr) except +
    long double modf(long double value, long double* iptr) except +
    float modff(float value, float* iptr) except +
    long double modfl(long double value, long double* iptr) except +

    float scalbn(float x, int n) except +
    double scalbn(double x, int n) except +
    long double scalbn(long double x, int n) except +
    float scalbnf(float x, int n) except +
    long double scalbnl(long double x, int n) except +

    float scalbln(float x, long int n) except +
    double scalbln(double x, long int n) except +
    long double scalbln(long double x, long int n) except +
    float scalblnf(float x, long int n) except +
    long double scalblnl(long double x, long int n) except +

    float cbrt(float x) except +
    double cbrt(double x) except +
    long double cbrt(long double x) except +
    float cbrtf(float x) except +
    long double cbrtl(long double x) except +

    # absolute values
    int abs(int j) except +
    long int abs(long int j) except +
    long long int abs(long long int j) except +
    float abs(float j) except +
    double abs(double j) except +
    long double abs(long double j) except +

    float fabs(float x) except +
    double fabs(double x) except +
    long double fabs(long double x) except +
    float fabsf(float x) except +
    long double fabsl(long double x) except +

    float hypot(float x, float y) except +
    double hypot(double x, double y) except +
    long double hypot(long double x, long double y) except +
    float hypotf(float x, float y) except +
    long double hypotl(long double x, long double y) except +

    # C++17 three-dimensional hypotenuse
    float hypot(float x, float y, float z) except +
    double hypot(double x, double y, double z) except +
    long double hypot(long double x, long double y, long double z) except +

    float pow(float x, float y) except +
    double pow(double x, double y) except +
    long double pow(long double x, long double y) except +
    float powf(float x, float y) except +
    long double powl(long double x, long double y) except +

    float sqrt(float x) except +
    double sqrt(double x) except +
    long double sqrt(long double x) except +
    float sqrtf(float x) except +
    long double sqrtl(long double x) except +

    float erf(float x) except +
    double erf(double x) except +
    long double erf(long double x) except +
    float erff(float x) except +
    long double erfl(long double x) except +

    float erfc(float x) except +
    double erfc(double x) except +
    long double erfc(long double x) except +
    float erfcf(float x) except +
    long double erfcl(long double x) except +

    float lgamma(float x) except +
    double lgamma(double x) except +
    long double lgamma(long double x) except +
    float lgammaf(float x) except +
    long double lgammal(long double x) except +

    float tgamma(float x) except +
    double tgamma(double x) except +
    long double tgamma(long double x) except +
    float tgammaf(float x) except +
    long double tgammal(long double x) except +

    float ceil(float x) except +
    double ceil(double x) except +
    long double ceil(long double x) except +
    float ceilf(float x) except +
    long double ceill(long double x) except +

    float floor(float x) except +
    double floor(double x) except +
    long double floor(long double x) except +
    float floorf(float x) except +
    long double floorl(long double x) except +

    float nearbyint(float x) except +
    double nearbyint(double x) except +
    long double nearbyint(long double x) except +
    float nearbyintf(float x) except +
    long double nearbyintl(long double x) except +

    float rint(float x) except +
    double rint(double x) except +
    long double rint(long double x) except +
    float rintf(float x) except +
    long double rintl(long double x) except +

    long int lrint(float x) except +
    long int lrint(double x) except +
    long int lrint(long double x) except +
    long int lrintf(float x) except +
    long int lrintl(long double x) except +

    long long int llrint(float x) except +
    long long int llrint(double x) except +
    long long int llrint(long double x) except +
    long long int llrintf(float x) except +
    long long int llrintl(long double x) except +

    float round(float x) except +
    double round(double x) except +
    long double round(long double x) except +
    float roundf(float x) except +
    long double roundl(long double x) except +

    long int lround(float x) except +
    long int lround(double x) except +
    long int lround(long double x) except +
    long int lroundf(float x) except +
    long int lroundl(long double x) except +

    long long int llround(float x) except +
    long long int llround(double x) except +
    long long int llround(long double x) except +
    long long int llroundf(float x) except +
    long long int llroundl(long double x) except +

    float trunc(float x) except +
    double trunc(double x) except +
    long double trunc(long double x) except +
    float truncf(float x) except +
    long double truncl(long double x) except +

    float fmod(float x, float y) except +
    double fmod(double x, double y) except +
    long double fmod(long double x, long double y) except +
    float fmodf(float x, float y) except +
    long double fmodl(long double x, long double y) except +

    float remainder(float x, float y) except +
    double remainder(double x, double y) except +
    long double remainder(long double x, long double y) except +
    float remainderf(float x, float y) except +
    long double remainderl(long double x, long double y) except +

    float remquo(float x, float y, int* quo) except +
    double remquo(double x, double y, int* quo) except +
    long double remquo(long double x, long double y, int* quo) except +
    float remquof(float x, float y, int* quo) except +
    long double remquol(long double x, long double y, int* quo) except +

    float copysign(float x, float y) except +
    double copysign(double x, double y) except +
    long double copysign(long double x, long double y) except +
    float copysignf(float x, float y) except +
    long double copysignl(long double x, long double y) except +

    double nan(const char* tagp) except +
    float nanf(const char* tagp) except +
    long double nanl(const char* tagp) except +

    float nextafter(float x, float y) except +
    double nextafter(double x, double y) except +
    long double nextafter(long double x, long double y) except +
    float nextafterf(float x, float y) except +
    long double nextafterl(long double x, long double y) except +

    float nexttoward(float x, long double y) except +
    double nexttoward(double x, long double y) except +
    long double nexttoward(long double x, long double y) except +
    float nexttowardf(float x, long double y) except +
    long double nexttowardl(long double x, long double y) except +

    float fdim(float x, float y) except +
    double fdim(double x, double y) except +
    long double fdim(long double x, long double y) except +
    float fdimf(float x, float y) except +
    long double fdiml(long double x, long double y) except +

    float fmax(float x, float y) except +
    double fmax(double x, double y) except +
    long double fmax(long double x, long double y) except +
    float fmaxf(float x, float y) except +
    long double fmaxl(long double x, long double y) except +

    float fmin(float x, float y) except +
    double fmin(double x, double y) except +
    long double fmin(long double x, long double y) except +
    float fminf(float x, float y) except +
    long double fminl(long double x, long double y) except +

    float fma(float x, float y, float z) except +
    double fma(double x, double y, double z) except +
    long double fma(long double x, long double y, long double z) except +
    float fmaf(float x, float y, float z) except +
    long double fmal(long double x, long double y, long double z) except +

    # C++20 linear interpolation
    float lerp(float a, float b, float t)
    double lerp(double a, double b, double t)
    long double lerp(long double a, long double b, long double t)

    # classification / comparison functions
    int fpclassify(float x) except +
    int fpclassify(double x) except +
    int fpclassify(long double x) except +

    bint isfinite(float x) except +
    bint isfinite(double x) except +
    bint isfinite(long double x) except +

    bint isinf(float x) except +
    bint isinf(double x) except +
    bint isinf(long double x) except +

    bint isnan(float x) except +
    bint isnan(double x) except +
    bint isnan(long double x) except +

    bint isnormal(float x) except +
    bint isnormal(double x) except +
    bint isnormal(long double x) except +

    bint signbit(float x) except +
    bint signbit(double x) except +
    bint signbit(long double x) except +

    bint isgreater(float x, float y) except +
    bint isgreater(double x, double y) except +
    bint isgreater(long double x, long double y) except +

    bint isgreaterequal(float x, float y) except +
    bint isgreaterequal(double x, double y) except +
    bint isgreaterequal(long double x, long double y) except +

    bint isless(float x, float y) except +
    bint isless(double x, double y) except +
    bint isless(long double x, long double y) except +

    bint islessequal(float x, float y) except +
    bint islessequal(double x, double y) except +
    bint islessequal(long double x, long double y) except +

    bint islessgreater(float x, float y) except +
    bint islessgreater(double x, double y) except +
    bint islessgreater(long double x, long double y) except +

    bint isunordered(float x, float y) except +
    bint isunordered(double x, double y) except +
    bint isunordered(long double x, long double y) except +

    # C++17 mathematical special functions

    # associated Laguerre polynomials
    double       assoc_laguerre(unsigned int n, unsigned int m, double x) except +
    float        assoc_laguerref(unsigned int n, unsigned int m, float x) except +
    long double  assoc_laguerrel(unsigned int n, unsigned int m, long double x) except +

    # associated Legendre functions
    double       assoc_legendre(unsigned int l, unsigned int m, double x) except +
    float        assoc_legendref(unsigned int l, unsigned int m, float x) except +
    long double  assoc_legendrel(unsigned int l, unsigned int m, long double x) except +

    # beta function
    double       beta(double x, double y) except +
    float        betaf(float x, float y) except +
    long double  betal(long double x, long double y) except +

    # complete elliptic integral of the first kind
    double       comp_ellint_1(double k) except +
    float        comp_ellint_1f(float k) except +
    long double  comp_ellint_1l(long double k) except +

    # complete elliptic integral of the second kind
    double       comp_ellint_2(double k) except +
    float        comp_ellint_2f(float k) except +
    long double  comp_ellint_2l(long double k) except +

    # complete elliptic integral of the third kind
    double       comp_ellint_3(double k, double nu) except +
    float        comp_ellint_3f(float k, float nu) except +
    long double  comp_ellint_3l(long double k, long double nu) except +

    # regular modified cylindrical Bessel functions
    double       cyl_bessel_i(double nu, double x) except +
    float        cyl_bessel_if(float nu, float x) except +
    long double  cyl_bessel_il(long double nu, long double x) except +

    # cylindrical Bessel functions of the first kind
    double       cyl_bessel_j(double nu, double x) except +
    float        cyl_bessel_jf(float nu, float x) except +
    long double  cyl_bessel_jl(long double nu, long double x) except +

    # irregular modified cylindrical Bessel functions
    double       cyl_bessel_k(double nu, double x) except +
    float        cyl_bessel_kf(float nu, float x) except +
    long double  cyl_bessel_kl(long double nu, long double x) except +

    # cylindrical Neumann functions
    # cylindrical Bessel functions of the second kind
    double       cyl_neumann(double nu, double x) except +
    float        cyl_neumannf(float nu, float x) except +
    long double  cyl_neumannl(long double nu, long double x) except +

    # incomplete elliptic integral of the first kind
    double       ellint_1(double k, double phi) except +
    float        ellint_1f(float k, float phi) except +
    long double  ellint_1l(long double k, long double phi) except +

    # incomplete elliptic integral of the second kind
    double       ellint_2(double k, double phi) except +
    float        ellint_2f(float k, float phi) except +
    long double  ellint_2l(long double k, long double phi) except +

    # incomplete elliptic integral of the third kind
    double       ellint_3(double k, double nu, double phi) except +
    float        ellint_3f(float k, float nu, float phi) except +
    long double  ellint_3l(long double k, long double nu, long double phi) except +

    # exponential integral
    double       expint(double x) except +
    float        expintf(float x) except +
    long double  expintl(long double x) except +

    # Hermite polynomials
    double       hermite(unsigned int n, double x) except +
    float        hermitef(unsigned int n, float x) except +
    long double  hermitel(unsigned int n, long double x) except +

    # Laguerre polynomials
    double       laguerre(unsigned int n, double x) except +
    float        laguerref(unsigned int n, float x) except +
    long double  laguerrel(unsigned int n, long double x) except +

    # Legendre polynomials
    double       legendre(unsigned int l, double x) except +
    float        legendref(unsigned int l, float x) except +
    long double  legendrel(unsigned int l, long double x) except +

    # Riemann zeta function
    double       riemann_zeta(double x) except +
    float        riemann_zetaf(float x) except +
    long double  riemann_zetal(long double x) except +

    # spherical Bessel functions of the first kind
    double       sph_bessel(unsigned int n, double x) except +
    float        sph_besself(unsigned int n, float x) except +
    long double  sph_bessell(unsigned int n, long double x) except +

    # spherical associated Legendre functions
    double       sph_legendre(unsigned int l, unsigned int m, double theta) except +
    float        sph_legendref(unsigned int l, unsigned int m, float theta) except +
    long double  sph_legendrel(unsigned int l, unsigned int m, long double theta) except +

    # spherical Neumann functions
    # spherical Bessel functions of the second kind
    double       sph_neumann(unsigned int n, double x) except +
    float        sph_neumannf(unsigned int n, float x) except +
    long double  sph_neumannl(unsigned int n, long double x) except +



cdef extern from "<limits>" namespace "std" nogil:
    enum float_round_style:
        round_indeterminate       = -1
        round_toward_zero         = 0
        round_to_nearest          = 1
        round_toward_infinity     = 2
        round_toward_neg_infinity = 3

    enum float_denorm_style:
        denorm_indeterminate  = -1
        denorm_absent         = 0
        denorm_present        = 1

    #The static methods can be called as, e.g. numeric_limits[int].round_error(), etc.
    #The const data members should be declared as static.  Cython currently doesn't allow that
    #and/or I can't figure it out, so you must instantiate an object to access, e.g.
    #cdef numeric_limits[double] lm
    #print lm.round_style
    cdef cppclass numeric_limits[T]:
        const bint is_specialized
        @staticmethod
        T min()
        @staticmethod
        T max()
        const int digits
        const int  digits10
        const bint is_signed
        const bint is_integer
        const bint is_exact
        const int radix
        @staticmethod
        T epsilon()
        @staticmethod
        T round_error()

        const int  min_exponent
        const int  min_exponent10
        const int  max_exponent
        const int  max_exponent10

        const bint has_infinity
        const bint has_quiet_NaN
        const bint has_signaling_NaN
        const float_denorm_style has_denorm
        const bint has_denorm_loss
        @staticmethod
        T infinity()
        @staticmethod
        T quiet_NaN()
        @staticmethod
        T signaling_NaN()
        @staticmethod
        T denorm_min()

        const bint is_iec559
        const bint is_bounded
        const bint is_modulo

        const bint traps
        const bint tinyness_before
        const float_round_style round_style



# cython: auto_pickle=False
"""
Plex - Transition Maps

This version represents state sets directly as dicts for speed.
"""

maxint = 2**31-1  # sentinel value


class TransitionMap(object):
    """
    A TransitionMap maps an input event to a set of states.
    An input event is one of: a range of character codes,
    the empty string (representing an epsilon move), or one
    of the special symbols BOL, EOL, EOF.

    For characters, this implementation compactly represents
    the map by means of a list:

      [code_0, states_0, code_1, states_1, code_2, states_2,
        ..., code_n-1, states_n-1, code_n]

    where |code_i| is a character code, and |states_i| is a
    set of states corresponding to characters with codes |c|
    in the range |code_i| <= |c| <= |code_i+1|.

    The following invariants hold:
      n >= 1
      code_0 == -maxint
      code_n == maxint
      code_i < code_i+1 for i in 0..n-1
      states_0 == states_n-1

    Mappings for the special events '', BOL, EOL, EOF are
    kept separately in a dictionary.
    """

    def __init__(self, map=None, special=None):
        if not map:
            map = [-maxint, {}, maxint]
        if not special:
            special = {}
        self.map = map          # The list of codes and states
        self.special = special  # Mapping for special events

    def add(self, event, new_state):
        """
        Add transition to |new_state| on |event|.
        """
        if type(event) is tuple:
            code0, code1 = event
            i = self.split(code0)
            j = self.split(code1)
            map = self.map
            while i < j:
                map[i + 1][new_state] = 1
                i += 2
        else:
            self.get_special(event)[new_state] = 1

    def add_set(self, event, new_set):
        """
        Add transitions to the states in |new_set| on |event|.
        """
        if type(event) is tuple:
            code0, code1 = event
            i = self.split(code0)
            j = self.split(code1)
            map = self.map
            while i < j:
                map[i + 1].update(new_set)
                i += 2
        else:
            self.get_special(event).update(new_set)

    def get_epsilon(self):
        """
        Return the mapping for epsilon, or None.
        """
        return self.special.get('')

    def iteritems(self):
        """
        Return the mapping as an iterable of ((code1, code2), state_set) and
        (special_event, state_set) pairs.
        """
        result = []
        map = self.map
        else_set = map[1]
        i = 0
        n = len(map) - 1
        code0 = map[0]
        while i < n:
            set = map[i + 1]
            code1 = map[i + 2]
            if set or else_set:
                result.append(((code0, code1), set))
            code0 = code1
            i += 2
        for event, set in self.special.items():
            if set:
                result.append((event, set))
        return iter(result)

    items = iteritems

    # ------------------- Private methods --------------------

    def split(self, code):
        """
        Search the list for the position of the split point for |code|,
        inserting a new split point if necessary. Returns index |i| such
        that |code| == |map[i]|.
        """
        # We use a funky variation on binary search.
        map = self.map
        hi = len(map) - 1
        # Special case: code == map[-1]
        if code == maxint:
            return hi

        # General case
        lo = 0
        # loop invariant: map[lo] <= code < map[hi] and hi - lo >= 2
        while hi - lo >= 4:
            # Find midpoint truncated to even index
            mid = ((lo + hi) // 2) & ~1
            if code < map[mid]:
                hi = mid
            else:
                lo = mid
        # map[lo] <= code < map[hi] and hi - lo == 2
        if map[lo] == code:
            return lo
        else:
            map[hi:hi] = [code, map[hi - 1].copy()]
            return hi

    def get_special(self, event):
        """
        Get state set for special event, adding a new entry if necessary.
        """
        special = self.special
        set = special.get(event, None)
        if not set:
            set = {}
            special[event] = set
        return set

    # --------------------- Conversion methods -----------------------

    def __str__(self):
        map_strs = []
        map = self.map
        n = len(map)
        i = 0
        while i < n:
            code = map[i]
            if code == -maxint:
                code_str = "-inf"
            elif code == maxint:
                code_str = "inf"
            else:
                code_str = str(code)
            map_strs.append(code_str)
            i += 1
            if i < n:
                map_strs.append(state_set_str(map[i]))
            i += 1
        special_strs = {}
        for event, set in self.special.items():
            special_strs[event] = state_set_str(set)
        return "[%s]+%s" % (
            ','.join(map_strs),
            special_strs
        )

    # --------------------- Debugging methods -----------------------

    def check(self):
        """Check data structure integrity."""
        if not self.map[-3] < self.map[-1]:
            print(self)
            assert 0

    def dump(self, file):
        map = self.map
        i = 0
        n = len(map) - 1
        while i < n:
            self.dump_range(map[i], map[i + 2], map[i + 1], file)
            i += 2
        for event, set in self.special.items():
            if set:
                if not event:
                    event = 'empty'
                self.dump_trans(event, set, file)

    def dump_range(self, code0, code1, set, file):
        if set:
            if code0 == -maxint:
                if code1 == maxint:
                    k = "any"
                else:
                    k = "< %s" % self.dump_char(code1)
            elif code1 == maxint:
                k = "> %s" % self.dump_char(code0 - 1)
            elif code0 == code1 - 1:
                k = self.dump_char(code0)
            else:
                k = "%s..%s" % (self.dump_char(code0),
                                self.dump_char(code1 - 1))
            self.dump_trans(k, set, file)

    def dump_char(self, code):
        if 0 <= code <= 255:
            return repr(chr(code))
        else:
            return "chr(%d)" % code

    def dump_trans(self, key, set, file):
        file.write("      %s --> %s\n" % (key, self.dump_set(set)))

    def dump_set(self, set):
        return state_set_str(set)


#
#   State set manipulation functions
#

def state_set_str(set):
    return "[%s]" % ','.join(["S%d" % state.number for state in set])



import unittest

from Cython.Utils import (
    _CACHE_NAME_PATTERN, _build_cache_name, _find_cache_attributes,
    build_hex_version, cached_method, clear_method_caches, try_finally_contextmanager)

METHOD_NAME = "cached_next"
CACHE_NAME = _build_cache_name(METHOD_NAME)
NAMES = CACHE_NAME, METHOD_NAME

class Cached(object):
    @cached_method
    def cached_next(self, x):
        return next(x)


class TestCythonUtils(unittest.TestCase):
    def test_build_hex_version(self):
        self.assertEqual('0x001D00A1', build_hex_version('0.29a1'))
        self.assertEqual('0x001D03C4', build_hex_version('0.29.3rc4'))
        self.assertEqual('0x001D00F0', build_hex_version('0.29'))
        self.assertEqual('0x040000F0', build_hex_version('4.0'))

    ############################## Cached Methods ##############################

    def test_cache_method_name(self):
        method_name = "foo"
        cache_name = _build_cache_name(method_name)
        match = _CACHE_NAME_PATTERN.match(cache_name)

        self.assertIsNot(match, None)
        self.assertEqual(match.group(1), method_name)

    def test_requirements_for_Cached(self):
        obj = Cached()

        self.assertFalse(hasattr(obj, CACHE_NAME))
        self.assertTrue(hasattr(obj, METHOD_NAME))
        self.set_of_names_equal(obj, set())

    def set_of_names_equal(self, obj, value):
        self.assertEqual(set(_find_cache_attributes(obj)), value)

    def test_find_cache_attributes(self):
        obj = Cached()
        method_name = "bar"
        cache_name = _build_cache_name(method_name)

        setattr(obj, CACHE_NAME, {})
        setattr(obj, cache_name, {})

        self.assertFalse(hasattr(obj, method_name))
        self.set_of_names_equal(obj, {NAMES, (cache_name, method_name)})

    def test_cached_method(self):
        obj = Cached()
        value = iter(range(3))  # iter for Py2
        cache = {(value,): 0}

        # cache args
        self.assertEqual(obj.cached_next(value), 0)
        self.set_of_names_equal(obj, {NAMES})
        self.assertEqual(getattr(obj, CACHE_NAME), cache)

        # use cache
        self.assertEqual(obj.cached_next(value), 0)
        self.set_of_names_equal(obj, {NAMES})
        self.assertEqual(getattr(obj, CACHE_NAME), cache)

    def test_clear_method_caches(self):
        obj = Cached()
        value = iter(range(3))  # iter for Py2
        cache = {(value,): 1}

        obj.cached_next(value)  # cache args

        clear_method_caches(obj)
        self.set_of_names_equal(obj, set())

        self.assertEqual(obj.cached_next(value), 1)
        self.set_of_names_equal(obj, {NAMES})
        self.assertEqual(getattr(obj, CACHE_NAME), cache)

    def test_clear_method_caches_with_missing_method(self):
        obj = Cached()
        method_name = "bar"
        cache_name = _build_cache_name(method_name)
        names = cache_name, method_name

        setattr(obj, cache_name, object())

        self.assertFalse(hasattr(obj, method_name))
        self.set_of_names_equal(obj, {names})

        clear_method_caches(obj)
        self.set_of_names_equal(obj, {names})

    def test_try_finally_contextmanager(self):
        states = []
        @try_finally_contextmanager
        def gen(*args, **kwargs):
            states.append("enter")
            yield (args, kwargs)
            states.append("exit")

        with gen(1, 2, 3, x=4) as call_args:
            assert states == ["enter"]
            self.assertEqual(call_args, ((1, 2, 3), {'x': 4}))
        assert states == ["enter", "exit"]

        class MyException(RuntimeError):
            pass

        del states[:]
        with self.assertRaises(MyException):
            with gen(1, 2, y=4) as call_args:
                assert states == ["enter"]
                self.assertEqual(call_args, ((1, 2), {'y': 4}))
                raise MyException("FAIL INSIDE")
            assert states == ["enter", "exit"]

        del states[:]
        with self.assertRaises(StopIteration):
            with gen(1, 2, y=4) as call_args:
                assert states == ["enter"]
                self.assertEqual(call_args, ((1, 2), {'y': 4}))
                raise StopIteration("STOP")
            assert states == ["enter", "exit"]
